%Script that builds the model based on parallax values
%Inputs is parallax values, row and column range in the area that has been mathces and the coordinates of the images
%Outputs are the X, Y and Z values of the model


% When calculating z values this code assumes rectified images with 
% shifting so that the difference in the corresponding image points is 
% equal only to the parallax and not the parallax + baseline. 

% If rectified images have not been shifted then the second term(b*scale) of 
% calculating the parallax in meters must be removed. 

function [X, Y, Z]  = make_model(parallax, row_range, col_range, XI1, YI1)

b = 0.10;                                                                   %Baseline in meters
pix_meters = 5.3 * 10^-6;                                                   %Pixels to meters
foc_length = 0.035;                                                         %Focal length in meters
height = 0.3;                                                               %Distance between object plane and camera
percent = 13;                                                                %Acceptable percentage devation of z-values around the difference between height and focal length

scale = foc_length/height;                                                  %Scaling factor in image

parallax_meters = parallax.*pix_meters./scale + b*scale;                    %Convert parallax in pixels to meters

parallax_meters(parallax_meters==b*scale) = 0;

Z_meters= foc_length - foc_length*b./parallax_meters;

%Remove extreme values
Z_meters(Z_meters<-(height-foc_length)*(1+percent/100)) = 0;
Z_meters(Z_meters>-(height-foc_length)*(1-percent/100)) = 0;


%Interpolation
[r,c,p] = find(Z_meters);                                                   %Find non-zero elements
[r0,c0] = find(Z_meters == 0);                                              %Find zero elemnts


% Alternativ way if size(parallax) = size(master)
% [r,c,p] = find(Z_meters(row_range(1):row_range(2),col_range(1):col_range(2)));                                                   
% r = r + row_range(1)-1;
% c = c + col_range(1)-1;
% 
% [r0,c0] = find(Z_meters(row_range(1):row_range(2),col_range(1):col_range(2)) == 0);                                              
% r0 = r0 + row_range(1)-1;
% c0 = c0 + col_range(1)-1;


F = TriScatteredInterp(r,c,p);                                              %Make interpolation model out of non-zero elements
Z_intp = F(r0,c0);                                                          %Interpolate the zero-elements


Z_intp=[p;Z_intp];
X=[r;r0];
Y=[c;c0];


Z_new = zeros(size(Z_meters));

%Arrange Z-values correct
for i=1:length(Z_intp)
    Z_new(X(i), Y(i)) = Z_intp(i);
end


Z = Z_new + height - foc_length;

X1_meters =... 
    (XI1(row_range(1):row_range(2),col_range(1):col_range(2))).*pix_meters;
Y1_meters =... 
    (YI1(row_range(1):row_range(2),col_range(1):col_range(2))).*pix_meters;


X = X1_meters.*(1-(Z_new./foc_length));
Y = Y1_meters.*(1-(Z_new./foc_length)); 

figure (101)
surf(100*X,100*Y,100*Z)
axis equal
colormap(jet)
    
%axis([X(1)*100 X(end)*100 Y(1)*100 Y(end)*100 -(height-foc_length)*100 -(height-foc_length)*93])
%axis([1 size(Z_new,2) 1 size(Z_new,1) 0 1.5])