% This function bring two images to epipolar orientation. Rectification is
% performed based on Computer Vision Toolbox. Rectified images can be used 
% for paralax calculation. 
% Input:image 1 undistorted and image 2 undistorted of the stereo images
% Output: image 1 in epipolar orientation and image 2 in epipolar orientation
% NB! Must have Computer vision Toolbox to run this script successfully


function [I1epi, I2epi] = rectify_images(I1u, I2u)
%% Step 1. Read Stereo Image Pair



cvexShowImagePair(uint8(I1u), uint8(I2u), 'Left image', 'Right image');
cvexShowMatches(uint8(I1u), uint8(I2u), [], [], 'Left image', 'Right image');

%% Step 2. Collect Salient Points from Each Image
close all

hcornerdet = vision.CornerDetector('Method','Local intensity comparison (Rosten & Drummond)','MaximumCornerCount', 1000, ...
    'CornerThreshold', 0.5, 'NeighborhoodSize', [21 21]);

cornerPoints1 = step(hcornerdet, I1u);

cornerPoints2 = step(hcornerdet, I2u);


% remove points on borders 
outliers1 = cornerPoints1(:,2) < 15;  
cornerPoints1(outliers1,:) = [];
outliers1 = cornerPoints1(:,2) > size(I1u,1)-15 ;
cornerPoints1(outliers1,:) = [];


outliers2 = cornerPoints2(:,2) < 15;  
cornerPoints2(outliers2,:) = [];
outliers2 = cornerPoints2(:,2) > size(I2u,1)-15 ;
cornerPoints2(outliers2,:) = [];



outliers1 = cornerPoints1(:,1) < 15;       % Pixels on white reference must be excluded from computations due to bad transformation estimation  
cornerPoints1(outliers1,:) = [];
outliers1 = cornerPoints1(:,1) > size(I1u,2)-15 ;
cornerPoints1(outliers1,:) = [];


outliers2 = cornerPoints2(:,1) < 15;  
cornerPoints2(outliers2,:) = [];
outliers2 = cornerPoints2(:,1) > size(I2u,2)-15 ;
cornerPoints2(outliers2,:) = [];

cvexShowImagePair(uint8(I1u), uint8(I2u), 'Corners in left image', ...
    'Corners in right image', 'SingleColor', cornerPoints1, cornerPoints2);

%% Step 3. Select Correspondences Between Points
close all
[features1, validPoints1] = extractFeatures(I1u, cornerPoints1, ...
    'BlockSize', 9);
[features2, validPoints2] = extractFeatures(I2u, cornerPoints2, ...
    'BlockSize', 9);

[indexPairs, match_metric] = matchFeatures(features1, features2, ...
    'Metric', 'normxcorr', 'MatchThreshold', 10);

% Retrieve locations of matched points for each image
matchedPoints1 = validPoints1(indexPairs(:, 1), :);
matchedPoints2 = validPoints2(indexPairs(:, 2), :);

cvexShowMatches(uint8(I1u), uint8(I2u), matchedPoints1, matchedPoints2, ...
  'Matched points in I1u', 'Matched points in I2u');

% avoid matching in Y direction
checkY = (double(matchedPoints2(:,2))-double(matchedPoints1(:,2)));
checkX = (double(matchedPoints2(:,1))-double(matchedPoints1(:,1)));

alf = atand(checkY./checkX);
m = mode(alf);

if m < 0

    outliers1 = alf < 2*m | alf > -2*m;
else
    outliers1 = alf > 2*m | alf < -2*m;
end
matchedPoints1(outliers1,:) = [];
matchedPoints2(outliers1,:) = [];
cvexShowMatches(uint8(I1u), uint8(I2u), matchedPoints1, matchedPoints2, ...
  'Matched points in I1u', 'Matched points in I2u');

            % matchedPoints1(outliers1,:) = [];
            % matchedPoints2(outliers1,:) = [];
            % 
            % checkY = abs(double(matchedPoints1(:,2))-double(matchedPoints2(:,2)));
            % outliers1 = (abs(checkY(:,1)) > 20);


% avoid long matching in X direction (RANSAC algoritgm does not perform
% well for 3D objects. Thus big outliers should be removed before running
% RANSAC)

dst(:,1) = sqrt((double(matchedPoints1(:,1))-double(matchedPoints2(:,1))).^2+(double(matchedPoints1(:,2))-double(matchedPoints2(:,2))).^2);
m = mode(dst)
outliers2 = dst < m-15 | dst > m+15;
matchedPoints1(outliers2,:) = [];
matchedPoints2(outliers2,:) = [];

            % sd = 50;
            % while sd >= 50
            % dst(:,1) = sqrt((double(matchedPoints1(:,1))-double(matchedPoints2(:,1))).^2+(double(matchedPoints1(:,2))-double(matchedPoints2(:,2))).^2);
            % dst(:,2) = mean(dst(:,1));
            % dst(:,3) = std(dst(:,1));
            % outliers2 = abs(dst(:,1) - dst(:,2)) > dst(:,3);
            % sd = dst(1,3)
            % clear('checkY','dst','outliers1','outliers2')
            % 
            % end
cvexShowMatches(uint8(I1u), uint8(I2u), matchedPoints1, matchedPoints2, ...
  'Matched points in I1u', 'Matched points in I2u');

cvexShowImagePair(uint8(I1u), uint8(I2u), 'Corners in left image', ...
    'Corners in right image', 'SingleColor', matchedPoints1, matchedPoints2);


%% Step 4. Remove Outliers Using Random Sample Consensus (RANSAC) Algorithm
close all
[fMatrix, inliers] = estimateFundamentalMatrix(...
    matchedPoints1, matchedPoints2, 'Method', 'RANSAC', ...
    'NumTrials', 1000, 'DistanceThreshold', 0.01, 'Confidence', 99.99);

inlierPoints1 = matchedPoints1(inliers, :);
inlierPoints2 = matchedPoints2(inliers, :);

cvexShowMatches(uint8(I1u), uint8(I2u), inlierPoints1, inlierPoints2, ...
    'Inlier points in I1u', 'Inlier points in I2u');



%% Step 5. Rectify Images (only for display)
close all
if isEpipoleInImage(fMatrix, size(I1u)) ...
    || isEpipoleInImage(fMatrix', size(I2u))
    error(['For the rectification to succeed, the epipoles must be ',...
        'outside the images.']);
end

[t1, t2] = estimateUncalibratedRectification(fMatrix, ...
    inlierPoints1, inlierPoints2, size(I2u));

cvexShowMatches(uint8(I1u), uint8(I2u), inlierPoints1, inlierPoints2, ...
    'Inlier points in I1u', 'Inlier points in I2u', 'RC', t1, t2);

cvexShowStereoImages('Rectified image 1', 'Rectified image 2',...
    uint8(I1u), uint8(I2u), inlierPoints1, inlierPoints2, fMatrix, t1, t2);

%% Step 6. Move images to epipolar orientation
close all

H1 = vision.GeometricTransformer(...
           'TransformMatrixSource', 'Property', ...
           'TransformMatrix',t1,...
           'OutputImagePositionSource', 'Property',...
           'OutputImagePosition', [-200 -200 1480 1224]);   % -200, -200 is used to acces image data after transformation

I1epi= step(H1,double(I1u));


H2 = vision.GeometricTransformer(...
           'TransformMatrixSource', 'Property', ...
           'TransformMatrix',t2,...
           'OutputImagePositionSource', 'Property',...
           'OutputImagePosition', [-200 -200 1480 1224]);

I2epi= step(H2,double(I2u));

figure
imagesc(I1epi)
colormap(gray)
figure
imagesc(I2epi)
colormap(gray)

%% find x0,y0 on rectified images

figure
imagesc(I1u)
colormap(gray)
hold on
plot(cc_1(2),cc_1(1),'.r')
plot(cc_1(2),cc_1(1),'or')


figure
imagesc(I1epi)
colormap(gray)
hold on
plot(cc_1(2),cc_1(1),'.r')
plot(cc_1(2),cc_1(1),'or')

x0 = round(cc_1(2));    % y coordinate --> N-S 
y0 = round(cc_1(1));    % x coordinate --> W-E

tempI1 = I1u(y0-15:y0+15,x0-15:x0+15);
R= zeros(size(I1epi));

for ii = x0-300:x0+300;
    for jj = y0-100:y0+100;
        tempI1epi = I1epi(jj-15:jj+15,ii-15:ii+15);
        R(ii,jj) = corr2(tempI1,tempI1epi);
    end
end

[C I] = max(abs(R(:)));
[rI1,cI1]= ind2sub(size(I2u),I);
plot(cI1,rI1,'.b');
plot(cI1,rI1,'ob');

clear('R', 'tempI1epi','tempI1');

figure
imagesc(I2u)
colormap(gray)
hold on
plot(cc_1(2),cc_1(1),'.r')
plot(cc_1(2),cc_1(1),'or')


figure
imagesc(I2epi)
colormap(gray)
hold on
plot(cc_1(2),cc_1(1),'.r')
plot(cc_1(2),cc_1(1),'or')


tempI2 = I2u(y0-15:y0+15,x0-15:x0+15);
R= zeros(size(I2epi));

for ii = x0-300:x0+300;
    for jj = y0-100:y0+100;
        tempI2epi = I2epi(jj-15:jj+15,ii-15:ii+15);
        R(ii,jj) = corr2(tempI2,tempI2epi);
    end
end

[C I] = max(abs(R(:)));
[rI2,cI2]= ind2sub(size(I2u),I);
plot(rI2,cI2,'.b')
plot(rI2,cI2,'ob')

% Calculate coordinates of pixels in new coordinate system
[X,Y]   = meshgrid(1:size(I1epi,2),1:size(I1epi,1));          % create coos matrix of distorted image


XI1 = X- rI1;
YI1 = Y- cI1;
XI2 = X- rI2;
YI2 = Y- cI2;

